// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ==  http://www.amdis-fem.org                                              ==
// ==                                                                        ==
// ============================================================================
//
// Software License for AMDiS
//
// Copyright (c) 2010 Dresden University of Technology 
// All rights reserved.
// Authors: Simon Vey, Thomas Witkowski et al.
//
// This file is part of AMDiS
//
// See also license.opensource.txt in the distribution.



/** \file RosenbrockStationary.h */

#ifndef AMDIS_ROSENBROCKSTATIONARY_H
#define AMDIS_ROSENBROCKSTATIONARY_H

#include "AMDiS_fwd.h"
#include "ProblemStat.h"
#include "SystemVector.h"
#include "time/RosenbrockMethod.h"

#ifdef HAVE_PARALLEL_DOMAIN_AMDIS
#ifdef HAVE_PARALLEL_MTL4
#include "parallel/Mtl4Solver.h"
#else
#include "parallel/PetscProblemStat.h"
#endif
#endif

namespace AMDiS {

  struct RosenbrockBoundary 
  {
    AbstractFunction<double, WorldVector<double> > *fct;
    
    DOFVector<double> *vec;

    int row;
    
    int col;
  };


  class RosenbrockStationary : public ProblemStat
  {
  public:
    RosenbrockStationary(std::string name)
      : ProblemStat(name),
	first(true),
	minusOne(-1.0),
	tauPtr(NULL),
	tauGamma(NULL),
	minusTauGamma(NULL),	
	invTauGamma(NULL),
	minusInvTauGamma(NULL)
    {}

    DOFVector<double>* getUnVec(int i)
    {
      return unVec->getDOFVector(i);
    }

    DOFVector<double>* getStageSolution(int i)
    {
      return stageSolution->getDOFVector(i);
    }

    DOFVector<double>* getTimeRhsVec(int i)
    {
      return timeRhsVec->getDOFVector(i);
    }

    void acceptTimestep();

    void buildAfterCoarsen(AdaptInfo *adaptInfo, Flag flag,
			   bool asmMatrix, bool asmVector);

    void solve(AdaptInfo *adaptInfo, bool, bool);

    void setRosenbrockMethod(RosenbrockMethod *method)
    {
      rm = method;
      init();
    }

    void reset()
    {
      first = true;
      stageSolution->set(0.0);
      unVec->set(0.0);
      for (int i = 0; i < rm->getStages(); i++)
	stageSolutions[i]->set(0.0);
    }

    void addOperator(Operator &op, int row, int col, 
		     double *factor = NULL, double *estFactor = NULL);

    void addJacobianOperator(Operator &op, int row, int col, 
			     double *factor = NULL, double *estFactor = NULL);

    void addTimeOperator(int i, int j);

    void setTau(double *ptr)
    {
      tauPtr = ptr;
    }

    void setTauGamma(double *ptr0, double *ptr1, double *ptr2, double *ptr3)
    {
      tauGamma = ptr0;
      minusTauGamma = ptr1;
      invTauGamma = ptr2;
      minusInvTauGamma = ptr3;
    }    

    double* getTauGamma()
    {
      return tauGamma;
    }

     double* getMinusTauGamma()
    {
      return minusTauGamma;
    }

    double* getInvTauGamma()
    {
      return invTauGamma;
    }

    double* getMinusInvTauGamma()
    {
      return minusInvTauGamma;
    }

   /// Adds a Dirichlet boundary condition, where the rhs is given by an 
    /// abstract function.
    void addDirichletBC(BoundaryType type, int row, int col,
			AbstractFunction<double, WorldVector<double> > *b);

    /// Adds a Dirichlet boundary condition, where the rhs is given by a DOF
    /// vector.
    void addDirichletBC(BoundaryType type, int row, int col,
			DOFVector<double> *vec)
    {
      FUNCNAME("RosenbrockStationary::addDirichletBC()");

      ERROR_EXIT("Not yet supported!\n");
    }

    /// Adds a Neumann boundary condition, where the rhs is given by an
    /// abstract function.
    void addNeumannBC(BoundaryType type, int row, int col, 
		      AbstractFunction<double, WorldVector<double> > *n)    
    {
      FUNCNAME("RosenbrockStationary::addNeumannBC()");

      ERROR_EXIT("Not yet supported!\n");    
    }

    /// Adds a Neumann boundary condition, where the rhs is given by an DOF
    /// vector.
    void addNeumannBC(BoundaryType type, int row, int col, 
		      DOFVector<double> *n)
    {
      FUNCNAME("RosenbrockStationary::addNeumannBC()");

      ERROR_EXIT("Not yet supported!\n");
    }

    /// Adds Robin boundary condition.
    void addRobinBC(BoundaryType type, int row, int col, 
		    AbstractFunction<double, WorldVector<double> > *n,
		    AbstractFunction<double, WorldVector<double> > *r)
    {
      FUNCNAME("RosenbrockStationary::addRobinBC()");

      ERROR_EXIT("Not yet supported!\n");
    }

#if 0
    /// Adds a periodic boundary condition.
    void addPeriodicBC(BoundaryType type, int row, int col)
    {
      FUNCNAME("RosenbrockStationary::addPeriodicBC()");

      ERROR_EXIT("Not yet supported!\n");
    }
#endif

  protected:
    void init();

  protected:    
    RosenbrockMethod *rm;

    SystemVector *stageSolution, *unVec, *timeRhsVec, *newUn, *tmp, *lowSol;

    std::vector<SystemVector*> stageSolutions;

    bool first;

    double minusOne;

    double *tauPtr;

    double *tauGamma, *minusTauGamma, *invTauGamma, *minusInvTauGamma;

    std::vector<RosenbrockBoundary> boundaries;
  };

}

#endif // AMDIS_ROSENBROCKSTATIONARY_H
